Require Import String Arith Program.
Require Import basis vgtac lang langlemmas.
Require Import simul_def.

Set Implicit Arguments.

(******************************
   similarity
 ******************************)

CoInductive similarity (e1 e2: exp) : Prop :=
| simil_fold (SIM: sim similarity e1 e2)
.

(**********************************************
   Verification using cofix
 **********************************************)

Tactic Notation "csimul_step" int_or_var(n) := apply simil_fold; simul_step n.

Lemma readsump_simulated_with_cofix: 
  forall n m (EQ: n = 2 * m), similarity (f @ n) (g @ m).
Proof.
  cofix CIH.
  intros; subst; do 6 csimul_step 1; do 2 csimul_step 0.
  destruct m0; csimul_step 2; [|by eauto]. 
  fold_rfg; generalize (m+(m+0)).
  cofix CIH'; intros.
  do 3 csimul_step 1.
  destruct n; csimul_step 1; [by eauto|].
  do 3 csimul_step 1; eauto.
Qed.

(**********************************************
   Decomposition of the verification
   The lemma should be defined transparently.
 **********************************************)

Lemma restart_simulated: forall (f g: voe) (n m: nat)
    (ASM: similarity ((e_of f) @ 0) ((e_of g) @ 0))
    (CLOf: fv_exp (e_of f) = []) (CLOg: fv_exp (e_of g) = []) (EQ: n = m),
  similarity (restart @ (e_of f) @ n) (restart @ (e_of g) @ m).
Proof.
  intros; simplcanon; revert m CLOf CLOg; cofix CIH; i.
  do 3 csimul_step 1.
  destruct ble_nat; csimul_step 1.
  - by rewrite !subst_exp_closed; eauto.
  - do 3 csimul_step 1.
    rewrite !subst_exp_closed; eauto.
Defined.  (* Must be transparent for the following proof via cofix. *)

Lemma fg_simulated_comp: forall n m (EQ: n = 2 * m),
  similarity (f @ n) (g @ m).
Proof.
  cofix CIH; i; subst.
  do 6 csimul_step 1; do 2 csimul_step 0.
  destruct m0; csimul_step 2; [|by eauto]. 
  eapply restart_simulated; s; eauto.
Qed. (* takes long to check guardedness *)

