Require Import String Arith Program.
Require Import basis vgtac lang langlemmas.
Require Import paco.
Require Import simul_def.

Set Implicit Arguments.

(*
   Note:
   We use the internal approach library
*)

(**********************
   Sound and Respectful up-to functions
 **********************)

Inductive soundclo (clo: exprel -> exprel) : Prop :=
_soundclo 
  (MON : monotone2 clo)
  (SOUND : forall r (PFIX: r <2= sim (clo r)), 
           r <2= paco2 sim bot2).
Hint Constructors soundclo.

Structure respectfulclo (clo: exprel -> exprel) : Prop :=
_respectfulclo 
  (MON: monotone2 clo)
  (RESPECT: forall l r (LE: l <2= r) (SIM: l <2= sim r),
            clo l <2= sim (clo r)).
Hint Constructors respectfulclo.

Inductive gresclo (r: exprel) (e1 e2: exp) : Prop := _gresclo
  clo (RES: respectfulclo clo) (CLO: clo r e1 e2).  
Hint Constructors gresclo.

Lemma sim_mon: monotone2 sim.
Proof.
  r; i; destruct IN; econs; eauto; i; exploit EXP; i; des; eauto 10.
Qed.

Lemma simclo_mon: forall clo, soundclo clo -> monotone2 (sim <*> clo).
Proof. intros; destruct H; eauto using sim_mon. Qed.
Hint Resolve simclo_mon : paco.

Lemma soundclo_is_sim: forall clo (UPTO: soundclo clo), 
  paco2 (sim <*> clo) bot2 <2= paco2 sim bot2.
Proof.
  i; punfold PR; edestruct UPTO.
  eapply (SOUND (paco2 (sim <*> clo) bot2)); eauto.
  i; punfold PR0. 
  by eapply simclo_mon; eauto; intros; destruct PR1; eauto.
Qed.

Lemma respectfulclo_is_sound: forall clo, respectfulclo clo -> soundclo clo.
Proof.
  set (rclo := fix rclo clo n (r: exprel) :=
    match n with
    | 0 => r
    | S n' => rclo clo n' r \2/ clo (rclo clo n' r)
    end).
  i; destruct H; econs; eauto.
  i; set (rr a b := exists n, rclo clo n r a b).
  assert (rr x0 x1) by (exists 0; eauto); clear PR.
  cut (forall n, rclo clo n r <2= sim (rclo clo (S n) r)).
  { intro X; revert x0 x1 H; pcofix CIH; i.
    unfold rr in *; des; eauto 10 using sim_mon. }
  induction n; i; [by s; eauto using sim_mon|].
  ss; des; [by eauto using sim_mon|]. 
  eapply sim_mon; [eapply RESPECT; [|apply IHn|]|]; inst; s; eauto.
Qed.

Lemma respectful_compose: forall clo1 clo2 
    (RES1: respectfulclo clo1) (RES2: respectfulclo clo2),
  respectfulclo (clo1 <*> clo2).
Proof. i; destruct RES1, RES2; eauto 10. Qed.

Lemma gresclo_respectful: respectfulclo gresclo.
Proof.
  econs; ii.
  - destructs IN RES; exists clo; eauto.
  - destructs PR RES; eapply sim_mon with (r:=clo r); eauto.
Qed.

Lemma simgres_mon: monotone2 (sim <*> gresclo).
Proof.
  destruct gresclo_respectful; eauto using sim_mon.
Qed.
Hint Resolve simgres_mon : paco.

Lemma gresclo_greatest: forall clo (RES: respectfulclo clo), clo <3= gresclo.
Proof.
  eauto.
Qed.

Lemma gresclo_incl: forall r, r <2= gresclo r.
Proof.
  i; eexists (fun x => x); eauto.
Qed.
Hint Resolve gresclo_incl.

Lemma gresclo_compose: forall clo (RES: respectfulclo clo) r,
  clo (gresclo r) <2= gresclo r.
Proof.
  i; eapply gresclo_greatest with (clo := clo <*> gresclo);
   eauto using respectful_compose, gresclo_respectful.
Qed.

Lemma gresclo_incl_rev: forall r,
  gresclo (paco2 (sim <*> gresclo) r) <2= paco2 (sim <*> gresclo) r.
Proof.
  intro r; pcofix CIH; i; pfold.
  eapply sim_mon, gresclo_compose, gresclo_respectful.
  destruct gresclo_respectful; eapply RESPECT, PR; i; [by eauto using gresclo_incl|].
  punfold PR0.
  by eapply simgres_mon; eauto; i; destruct PR1; eauto.
Qed.

(**********************
   Up-to context
 **********************)

(* up-to function for context closure *)
Inductive ctxclo (r: exprel) : exp -> exp -> Prop :=
| ctx_id: forall e, ctxclo r e e
| ctx_appv: forall f x e f' x' e' e2 e2'
              (R1: forall v: val, 
                   r (subst_exp [(x,v:exp); (f, Efix f x e)] e)
                     (subst_exp [(x',v:exp); (f', Efix f' x' e')] e'))
              (R2: r e2 e2'),
            ctxclo r (Eapp (Efix f x e) e2) (Eapp (Efix f' x' e') e2')
| ctx_seq: forall e1 e1' e2 e2' (R2: r e1 e1') (R2: r e2 e2'), 
           ctxclo r (Eseq e1 e2) (Eseq e1' e2')
| ctx_if: forall e e' e1 e1' e2 e2' (R: r e e') (R1: r e1 e1') (R2: r e2 e2'), 
          ctxclo r (Eif e e1 e2) (Eif e' e1' e2')
| ctx_tau: forall e e' e'' (RED: reduce e' Qtau e'') (R: r e e''), 
           ctxclo r e e'
| ctx_r: forall e e' (R: r e e'), ctxclo r e e'
.
Hint Constructors ctxclo.

Lemma ctxclo_is_respectful: respectfulclo ctxclo.
Proof.
  econs; [by ii; destruct IN; eauto|].
  ii; destruct PR.
  - (* syntactically identical *) 
    by econs; i; esplits; subst; eauto using taur_nil.
  - (* appv *)
    econs; i; [by simplcanon; destruct v; inv VAL|].
    apply SIM in R2; destruct R2.
    destruct (value_dec e2); des; subst.
    + simplcanon; eapply reduce_det in RED; [|step].
      s in RED; des; subst; esplits.
      * eauto using taureduces_scont.
      * s; eauto.
      * s; eauto.
    + eapply reduce_inv_scont in RED; eauto; des; subst.
      eapply EXP in RED; des; subst; esplits.
      * s; eauto using taureduces_scont.
      * s; eauto using qreduce_scont.
      * s; eauto.
  - (* seq *)
    econs; i; [by simplcanon; destruct v; inv VAL|].
    apply SIM in R2; destruct R2.
    destruct (value_dec e1); des; subst.
    + simplcanon; eapply reduce_det in RED; [|step].
      s in RED; des; subst; esplits.
      * eauto using taureduces_scont.
      * s; eauto.
      * s; eauto.
    + eapply reduce_inv_scont in RED; eauto; des; subst.
      eapply EXP in RED; des; subst; esplits.
      * eauto using taureduces_scont.
      * s; eauto using qreduce_scont.
      * s; eauto.
  - (* if *) 
    econs; i; [by simplcanon; destruct v; inv VAL|].
    apply SIM in R; destruct R.
    destruct (value_dec e); des; subst. 
    + depdes RED; simplcanon.
      * esplits; (inst; s; eauto using taureduces_scont).
      * esplits; (inst; s; eauto using taureduces_scont).
      * destruct SK; inv x; exfalso; eauto using val_nored.
    + eapply reduce_inv_scont in RED; eauto; des; subst. 
      eapply EXP in RED; des; subst; esplits.
      * eauto using taureduces_scont.
      * s; eauto using qreduce_scont.
      * s; eauto.
  - (* tau red *)
    apply SIM in R; destruct R; econs; eauto.
    i; exploit EXP; eauto.
    i; des; esplits; eauto.
  - (* r *)
    eauto using sim_mon.
Qed.

(**********************
   Verification
 **********************)

Tactic Notation "pusimil_step" int_or_var(n) := 
  try apply gresclo_incl; try left; pfold; intros; simul_step n.

Ltac appctx := apply (gresclo_compose ctxclo_is_respectful); econs.

Lemma restart_simulated: forall (f g: voe) (n m: nat) (r: exprel) 
    (ASM: r ((e_of f) @ 0) ((e_of g) @ 0))
    (CLOf: fv_exp (e_of f) = []) (CLOg: fv_exp (e_of g) = []) (EQ: n = m),
  paco2 (sim <*> gresclo) r (restart @ (e_of f) @ n) (restart @ (e_of g) @ m).
Proof.
  pcofix CIH; i; simplcanon.
  do 2 pusimil_step 1; rewrite !subst_exp_closed; eauto.
  appctx; [appctx|eauto|appctx; [appctx|]].
  apply gresclo_incl; pusimil_step 1; eauto.
Qed.

Lemma fg_simulated_using_lemma: forall n m (EQ: n = 2 * m),
  paco2 (sim <*> gresclo) bot2 (f @ n) (g @ m).
Proof.
  pcofix CIH; i; subst.
  do 6 pusimil_step 1; do 2 pusimil_step 0.
  destruct m0; pusimil_step 2; [|by eauto]. 
  apply gresclo_incl; left; eapply restart_simulated; s; eauto.
Qed.

Lemma fg_simulated: 
  forall n m (EQ: n = 2 * m), paco2 sim bot2 (f @ n) (g @ m).
Proof. 
  intros.
  eapply (soundclo_is_sim (respectfulclo_is_sound gresclo_respectful)).
  by apply fg_simulated_using_lemma. 
Qed.
